import copy
import os
from collections import namedtuple

from conans.model.ref import ConanFileReference
from conans.model.version import Version
from cpt.tools import split_colon_env, transform_list_options_to_dict

default_gcc_versions = ["4.9", "5", "6", "7", "8", "9", "10", "11"]
default_clang_versions = ["3.8", "3.9", "4.0", "5.0", "6.0", "7.0", "8", "9", "10", "11", "12", "13"]
default_visual_versions = ["14", "15", "16", "17"]
default_visual_runtimes = ["MT", "MD", "MTd", "MDd"]
default_msvc_versions = ["190", "191", "192", "193"]
default_msvc_runtimes = ["static", "dynamic"]
default_apple_clang_versions = ["9.0", "9.1", "10.0", "11.0", "12.0"]
default_archs = ["x86", "x86_64"]
default_build_types = ["Release", "Debug"]
default_cppstds = [None]


def get_mingw_package_reference():
    env_ref = os.getenv("CONAN_MINGW_INSTALLER_REFERENCE")
    return ConanFileReference.loads(env_ref or "mingw-w64/8.1")


def get_mingw_config_from_env():
    tmp = os.getenv("MINGW_CONFIGURATIONS", "")
    # 4.9@x86_64@seh@posix",4.9@x86_64@seh@win32"
    if not tmp:
        return []
    ret = []
    confs = tmp.split(",")
    for conf in confs:
        conf = conf.strip()
        ret.append(conf.split("@"))
    return ret


def get_env_visual_toolsets():
    ret = {}
    var = split_colon_env("CONAN_VISUAL_TOOLSETS")
    if not var:
        return None
    for toolset_versions in var:
        tmp = toolset_versions.split("=")
        if len(tmp) != 2:
            raise Exception("Invalid value for CONAN_VISUAL_TOOLSETS env variable, use "
                            "'14=v140;v140_xp,12=v120_xp'")
        visual_version, tmp_toolsets = tmp
        toolsets = tmp_toolsets.split(";")
        ret[visual_version] = toolsets
    return ret


class BuildGenerator(object):

    def __init__(self, reference, os_name, gcc_versions, apple_clang_versions, clang_versions,
                 visual_versions, visual_runtimes, visual_toolsets, vs10_x86_64_enabled,
                 msvc_versions, msvc_runtimes, msvc_runtime_types,
                 mingw_configurations, archs, allow_gcc_minors, build_types, options, cppstds):

        self._visual_toolsets = visual_toolsets
        self._os_name = os_name
        self._reference = reference
        self._vs10_x86_64_enabled = vs10_x86_64_enabled
        self._allow_gcc_minors = allow_gcc_minors or os.getenv("CONAN_ALLOW_GCC_MINORS", False)

        self._clang_versions = clang_versions or split_colon_env("CONAN_CLANG_VERSIONS")
        self._gcc_versions = gcc_versions or split_colon_env("CONAN_GCC_VERSIONS")

        # If there are some GCC versions declared then we don't default the clang
        # versions
        if self._clang_versions is None:
            if not self._gcc_versions:
                self._clang_versions = default_clang_versions
            else:
                self._clang_versions = []

        # If there are some CLANG versions declared then we don't default the gcc
        # versions
        if self._gcc_versions is None:
            if self._clang_versions == default_clang_versions:
                self._gcc_versions = default_gcc_versions
            else:
                self._gcc_versions = []

        if self._gcc_versions and not self._allow_gcc_minors:
            for a_version in self._gcc_versions:
                if Version(str(a_version)) >= Version("5") and "." in str(a_version):
                    raise Exception("""
******************* DEPRECATED GCC MINOR VERSIONS! ***************************************

- The use of gcc versions >= 5 and specifying the minor version (e.j "5.4") is deprecated.
- The ABI of gcc >= 5 (5, 6, and 7) is compatible between minor versions (e.j 5.3 is compatible with 5.4)
- Specify only the major in your script:
   - CONAN_GCC_VERSIONS="5,6,7" if you are using environment variables.
   - gcc_versions=["5", "6", "7"] if you are using the constructor parameter.

You can still keep using the same docker images, or use the new "conanio/gcc5", "conanio/gcc6", "conanio/gcc7"

If you still want to keep the old behavior, set the environment var CONAN_ALLOW_GCC_MINORS or pass the
"allow_gcc_minors=True" parameter. But it is not recommended, if your packages are public most users
won't be able to use them.

******************************************************************************************

""")

        self._msvc_versions = msvc_versions or split_colon_env("CONAN_MSVC_VERSIONS")
        self._visual_versions = visual_versions or split_colon_env("CONAN_VISUAL_VERSIONS")

        if self._msvc_versions is None:
            if not mingw_configurations and not get_mingw_config_from_env() and not self._visual_versions:
                self._msvc_versions = default_msvc_versions
            else:
                self._msvc_versions = []
        elif mingw_configurations or get_mingw_config_from_env():
            self._msvc_versions = []

        if self._msvc_versions is not None:
            self._msvc_runtimes = msvc_runtimes or split_colon_env("CONAN_MSVC_RUNTIMES")
            if self._msvc_runtimes is None:
                self._msvc_runtimes = default_msvc_runtimes

            self._msvc_runtime_types = msvc_runtime_types or split_colon_env("CONAN_MSVC_RUNTIME_TYPES")
            if self._msvc_runtime_types is None:
                self._msvc_runtime_types = default_build_types

        if self._visual_versions is None:
            if not mingw_configurations and not get_mingw_config_from_env() and not self._msvc_versions:
                self._visual_versions = default_visual_versions
            else:
                self._visual_versions = []
        elif mingw_configurations or get_mingw_config_from_env():
            self._visual_versions = []

        if self._visual_versions is not None:
            self._visual_runtimes = visual_runtimes or split_colon_env("CONAN_VISUAL_RUNTIMES")
            if self._visual_runtimes is None:
                self._visual_runtimes = default_visual_runtimes

        self._apple_clang_versions = apple_clang_versions or split_colon_env("CONAN_APPLE_CLANG_VERSIONS")

        if self._apple_clang_versions is None:
            self._apple_clang_versions = default_apple_clang_versions

        self._mingw_configurations = mingw_configurations or get_mingw_config_from_env()

        _default_archs = ["x86_64"] if self._os_name == "Darwin" else default_archs

        self._archs = archs or split_colon_env("CONAN_ARCHS")
        if self._archs is None:
            self._archs = _default_archs

        self._build_types = build_types or split_colon_env("CONAN_BUILD_TYPES")
        if self._build_types is None:
            self._build_types = default_build_types

        self._cppstds = cppstds or split_colon_env("CONAN_CPPSTDS")
        if self._cppstds is None:
            self._cppstds = default_cppstds

        options = options or split_colon_env("CONAN_OPTIONS")
        if options is None:
            options = []
        self._options = transform_list_options_to_dict(options)

    def get_builds(self, pure_c, shared_option_name, dll_with_static_runtime, reference=None,
                   build_all_options_values=None):

        ref = reference or self._reference

        if self._os_name == "Windows":
            if self._mingw_configurations:
                builds = get_mingw_builds(self._mingw_configurations,
                                          get_mingw_package_reference(), self._archs,
                                          shared_option_name, self._build_types, self._cppstds,
                                          self._options, ref, build_all_options_values)
            elif self._msvc_versions:
                builds = get_msvc_builds(self._msvc_versions, self._archs,
                                         self._msvc_runtimes, self._msvc_runtime_types,
                                         shared_option_name, dll_with_static_runtime, self._build_types,
                                         self._cppstds, self._options, ref,
                                         build_all_options_values)
            else:
                builds = []
            builds.extend(get_visual_builds(self._visual_versions, self._archs,
                                            self._visual_runtimes, self._visual_toolsets,
                                            shared_option_name, dll_with_static_runtime,
                                            self._vs10_x86_64_enabled,
                                            self._build_types, self._cppstds, self._options, ref,
                                            build_all_options_values))
            return builds
        elif self._os_name == "Linux":
            builds = get_linux_gcc_builds(self._gcc_versions, self._archs, shared_option_name,
                                          pure_c, self._build_types, self._cppstds, self._options, ref,
                                          build_all_options_values)
            builds.extend(get_linux_clang_builds(self._clang_versions, self._archs,
                                                 shared_option_name, pure_c, self._build_types,
                                                 self._cppstds, self._options, ref,
                                                 build_all_options_values))
            return builds
        elif self._os_name == "Darwin":
            return get_osx_apple_clang_builds(self._apple_clang_versions, self._archs,
                                              shared_option_name, pure_c, self._build_types,
                                              self._cppstds,  self._options, ref,
                                              build_all_options_values)
        elif self._os_name == "FreeBSD":
            return get_linux_clang_builds(self._clang_versions, self._archs, shared_option_name,
                                          pure_c, self._build_types, self._cppstds, self._options,
                                          ref, build_all_options_values)
        else:
            raise Exception("Unknown operating system: %s" % self._os_name)


class BuildConf(namedtuple("BuildConf", "settings options env_vars build_requires reference")):

    def __new__(cls, settings, options, env_vars, build_requires, reference):
        if not isinstance(settings, dict):
            raise Exception("'settings' field has to be a dict")
        if not isinstance(options, dict):
            raise Exception("'options' field has to be a dict")
        if not isinstance(env_vars, dict):
            raise Exception("'env_vars' field has to be a dict")
        if not isinstance(build_requires, dict):
            raise Exception("'build_requires' field has to be a dict")
        if reference is not None and not isinstance(reference, str) and \
           not isinstance(reference, ConanFileReference):
            raise Exception("'reference' field has to be a string or ConanFileReference")

        if isinstance(reference, str):
            reference = ConanFileReference.loads(reference)

        return super(BuildConf, cls).__new__(cls, settings, options, env_vars, build_requires,
                                             reference)


def get_mingw_builds(mingw_configurations, mingw_installer_reference,
                     archs, shared_option_name, build_types, cppstds,
                     options, reference=None, build_all_options_values=None):
    builds = []
    for config in mingw_configurations:
        version, arch, exception, thread = config
        if arch not in archs:
            continue
        settings = {"arch": arch, "compiler": "gcc",
                    "compiler.version": version[0:3],
                    "compiler.threads": thread,
                    "compiler.exception": exception}
        build_requires = {"*": [mingw_installer_reference]}

        if shared_option_name and not build_all_options_values:
            for shared in [False, True]:
                opt = copy.copy(options)
                opt[shared_option_name] = shared
                builds += _make_mingw_builds(settings, opt, build_requires, build_types, cppstds, reference)
        elif build_all_options_values:
            for option_values in build_all_options_values:
                opt = copy.copy(options)
                opt.update(option_values)
                builds += _make_mingw_builds(settings, opt, build_requires, build_types, cppstds, reference)
        else:
            builds += _make_mingw_builds(settings, copy.copy(options), build_requires, build_types, cppstds, reference)

    return builds


def _make_mingw_builds(settings, options, build_requires, build_types, cppstds, reference=None):
    builds = []

    for build_type_it in build_types:
        for cppstd in cppstds:
            s2 = copy.copy(settings)
            s2.update({"build_type": build_type_it})
            s2.update({"compiler.libcxx": "libstdc++"})
            if cppstd:
                s2.update({"compiler.cppstd": cppstd})
            builds.append(BuildConf(s2, options, {}, build_requires, reference))

    return builds


def get_visual_builds(visual_versions, archs, visual_runtimes, visual_toolsets, shared_option_name,
                      dll_with_static_runtime, vs10_x86_64_enabled, build_types, cppstds,
                      options, reference=None, build_all_options_values=None):

    visual_toolsets = visual_toolsets or get_env_visual_toolsets()
    ret = []
    for visual_version in visual_versions:
        visual_version = str(visual_version)
        for arch in archs:
            if not vs10_x86_64_enabled and arch == "x86_64" and visual_version == "10":
                continue
            if not visual_toolsets or not visual_toolsets.get(visual_version):
                toolsets = [None]
            else:
                toolsets = visual_toolsets.get(visual_version)
            for toolset in toolsets:
                visual_builds = get_visual_builds_for_version(visual_runtimes, visual_version, arch,
                                                              shared_option_name,
                                                              dll_with_static_runtime, build_types,
                                                              cppstds, options, reference,
                                                              toolset=toolset,
                                                              build_all_options_values=build_all_options_values)
                ret.extend(visual_builds)
    return ret


def get_visual_builds_for_version(visual_runtimes, visual_version, arch, shared_option_name,
                                  dll_with_static_runtime, build_types, cppstds, options,
                                  reference=None, toolset=None, build_all_options_values=None):
    base_set = {"compiler": "Visual Studio",
                "compiler.version": visual_version,
                "arch": arch}

    if toolset:
        base_set["compiler.toolset"] = toolset
    sets = []

    debug_builds = set(['Debug'])
    release_builds = set(['Release', 'RelWithDebInfo', 'MinSizeRel'])

    runtime_build_map = {
        'MTd': debug_builds,
        'MDd': debug_builds,
        'MT': release_builds,
        'MD': release_builds,
    }

    build_types_set = set(build_types)

    for rt in visual_runtimes:
        for cppstd in cppstds:
            compatible_builds = list(build_types_set.intersection(runtime_build_map[rt]))
            compatible_builds.sort() # so that it is deterministic.
            for bld in compatible_builds:
                partial_settings = {"build_type": bld, "compiler.runtime": rt}
                if cppstd:
                    partial_settings["compiler.cppstd"] = cppstd
                if shared_option_name and not build_all_options_values:
                    opt = copy.copy(options)
                    opt[shared_option_name] = False
                    sets.append((partial_settings, opt, {}, {}))
                    if rt in ['MT', 'MTd']:
                        if dll_with_static_runtime:
                            opt = copy.copy(options)
                            opt[shared_option_name] = True
                            sets.append((partial_settings,
                                        opt, {}, {}))
                    else:
                        opt = copy.copy(options)
                        opt[shared_option_name] = True
                        sets.append((partial_settings,
                                    opt, {}, {}))
                elif build_all_options_values:
                    for option_values in build_all_options_values:
                        opt = copy.copy(options)
                        opt.update(option_values)
                        sets.append((partial_settings, opt, {}, {}))
                        if shared_option_name and rt in ['MT', 'MTd'] and dll_with_static_runtime:
                            new_opt = copy.copy(opt)
                            new_opt[shared_option_name] = True
                            sets.append((partial_settings, opt, {}, {}))
                else:
                    sets.append((partial_settings, options, {}, {}))

    ret = []
    for setting, options, env_vars, build_requires in sets:
        tmp = copy.copy(base_set)
        tmp.update(setting)
        ret.append(BuildConf(tmp, copy.copy(options), env_vars, build_requires, reference))

    return ret


def get_msvc_builds(msvc_versions, archs, msvc_runtimes, msvc_runtime_types, shared_option_name,
                    dll_with_static_runtime, build_types, cppstds, options, reference=None,
                    build_all_options_values=None):

    msvc_runtime_types = msvc_runtime_types or split_colon_env("CONAN_BUILD_TYPES") or default_build_types
    ret = []
    for msvc_version in msvc_versions:
        msvc_version = str(msvc_version)
        for arch in archs:
            for runtime_type in msvc_runtime_types:
                msvc_builds = get_msvc_builds_for_version(msvc_runtimes, msvc_version, arch,
                                                          shared_option_name, dll_with_static_runtime, build_types,
                                                          cppstds, options, reference,
                                                          runtime_type=runtime_type,
                                                          build_all_options_values=build_all_options_values)
                ret.extend(msvc_builds)
    return ret


def get_msvc_builds_for_version(msvc_runtimes, msvc_version, arch, shared_option_name, dll_with_static_runtime,
                                build_types, cppstds, options, reference=None,
                                runtime_type=None, build_all_options_values=None):
    base_set = {"compiler": "msvc",
                "compiler.version": msvc_version,
                "arch": arch,
                "compiler.runtime_type": runtime_type}
    if runtime_type is None and msvc_runtimes:
        base_set["compiler.runtime_type"] = "Release"
        runtime_type = "Release"
    sets = []

    debug_builds = ['Debug']
    release_builds = ['Release', 'RelWithDebInfo', 'MinSizeRel']
    rtt_build_map = {"release": release_builds, "debug": debug_builds}
    build_types_set = set(build_types)

    for rt in msvc_runtimes:
        for cppstd in cppstds:
            compatible_builds = list(build_types_set.intersection(rtt_build_map[runtime_type.lower()]))
            compatible_builds.sort()
            for bld in compatible_builds:
                partial_settings = {"build_type": bld, "compiler.runtime": rt}
                if cppstd:
                    partial_settings["compiler.cppstd"] = cppstd
                if shared_option_name and not build_all_options_values:
                    opt = copy.copy(options)
                    opt[shared_option_name] = False
                    sets.append((partial_settings, opt, {}, {}))
                    if (rt == "static" and dll_with_static_runtime) or rt == "dynamic":
                        opt = copy.copy(options)
                        opt[shared_option_name] = True
                        sets.append((partial_settings, opt, {}, {}))
                elif build_all_options_values:
                    for option_values in build_all_options_values:
                        opt = copy.copy(options)
                        opt.update(option_values)
                        sets.append((partial_settings, opt, {}, {}))
                        if shared_option_name and rt == "static" and dll_with_static_runtime:
                            new_opt = copy.copy(opt)
                            new_opt[shared_option_name] = True
                            sets.append((partial_settings, opt, {}, {}))
                else:
                    sets.append((partial_settings, options, {}, {}))

    ret = []
    for setting, options, env_vars, build_requires in sets:
        tmp = copy.copy(base_set)
        tmp.update(setting)
        ret.append(BuildConf(tmp, copy.copy(options), env_vars, build_requires, reference))

    return ret


def get_build(compiler, the_arch, the_build_type, the_compiler_version,
              the_compiler_cppstd, the_libcxx, options, reference):
    setts = {"arch": the_arch,
             "build_type": the_build_type,
             "compiler": compiler,
             "compiler.version": the_compiler_version}
    if the_compiler_cppstd:
        setts["compiler.cppstd"] = the_compiler_cppstd
    if the_libcxx:
        setts["compiler.libcxx"] = the_libcxx

    return BuildConf(setts, copy.copy(options), {}, {}, reference)


def get_osx_apple_clang_builds(apple_clang_versions, archs, shared_option_name,
                               pure_c, build_types, cppstds, options, reference=None,
                               build_all_options_values=None):
    ret = []
    # Not specified compiler or compiler version, will use the auto detected
    for compiler_version in apple_clang_versions:
        for arch in archs:
            for cppstd in cppstds:
                if shared_option_name and not build_all_options_values:
                    for shared in [False, True]:
                        opt = copy.copy(options)
                        opt[shared_option_name] = shared
                        for build_type_it in build_types:
                            if not pure_c:
                                ret.append(get_build("apple-clang", arch, build_type_it,
                                                        compiler_version, cppstd,
                                                        "libc++", opt, reference))
                            else:
                                ret.append(get_build("apple-clang", arch, build_type_it,
                                                        compiler_version, None,
                                                        None, opt, reference))
                elif build_all_options_values:
                    for option_values in build_all_options_values:
                        opt = copy.copy(options)
                        opt.update(option_values)
                        for build_type_it in build_types:
                            if not pure_c:
                                ret.append(get_build("apple-clang", arch, build_type_it,
                                                     compiler_version, cppstd,
                                                     "libc++", opt, reference))
                            else:
                                ret.append(get_build("apple-clang", arch, build_type_it,
                                                     compiler_version, None,
                                                     None, opt, reference))
                else:
                    for build_type_it in build_types:
                        if not pure_c:
                            ret.append(get_build("apple-clang", arch, build_type_it,
                                                    compiler_version, cppstd,
                                                    "libc++", options, reference))
                        else:
                            ret.append(get_build("apple-clang", arch, build_type_it,
                                                    compiler_version, None,
                                                    None, options, reference))

    return ret


def get_linux_gcc_builds(gcc_versions, archs, shared_option_name, pure_c, build_types, cppstds,
                         options, reference=None, build_all_options_values=None):
    ret = []
    # Not specified compiler or compiler version, will use the auto detected
    for gcc_version in gcc_versions:
        for arch in archs:
            for cppstd in cppstds:
                if shared_option_name and not build_all_options_values:
                    for shared in [False, True]:
                        opt = copy.copy(options)
                        opt[shared_option_name] = shared
                        for build_type_it in build_types:
                            if not pure_c:
                                ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                     cppstd, "libstdc++", opt, reference))
                                if float(gcc_version) >= 5:
                                    ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                         cppstd, "libstdc++11", opt, reference))
                            else:
                                ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                     None, None, opt, reference))
                elif build_all_options_values:
                    for option_values in build_all_options_values:
                        opt = copy.copy(options)
                        opt.update(option_values)
                        for build_type_it in build_types:
                            if not pure_c:
                                ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                     cppstd, "libstdc++", opt, reference))
                                if float(gcc_version) >= 5:
                                    ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                         cppstd, "libstdc++11", opt, reference))
                            else:
                                ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                     None, None, opt, reference))

                else:
                    for build_type_it in build_types:
                        if not pure_c:
                            ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                 cppstd, "libstdc++", options, reference))
                            if float(gcc_version) >= 5:
                                ret.append(get_build("gcc", arch, build_type_it, gcc_version,
                                                     cppstd, "libstdc++11", options, reference))
                        else:
                            ret.append(get_build("gcc", arch, build_type_it, gcc_version, None,
                                                 None, options, reference))
    return ret


def get_linux_clang_builds(clang_versions, archs, shared_option_name, pure_c, build_types, cppstds,
                           options, reference=None, build_all_options_values=None):
    ret = []
    # Not specified compiler or compiler version, will use the auto detected
    for clang_version in clang_versions:
        for arch in archs:
            for cppstd in cppstds:
                if shared_option_name and not build_all_options_values:
                    for shared in [False, True]:
                        opt = copy.copy(options)
                        opt[shared_option_name] = shared
                        for build_type_it in build_types:
                            if not pure_c:
                                ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                     cppstd, "libstdc++", opt, reference))
                                ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                     cppstd, "libc++", opt, reference))
                            else:
                                ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                     None, None, opt, reference))
                elif build_all_options_values:
                    for option_values in build_all_options_values:
                        opt = copy.copy(options)
                        opt.update(option_values)
                        for build_type_it in build_types:
                            if not pure_c:
                                ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                     cppstd, "libstdc++", opt, reference))
                                ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                     cppstd, "libc++", opt, reference))
                            else:
                                ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                     None, None, opt, reference))
                else:
                    for build_type_it in build_types:
                        if not pure_c:
                            ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                 cppstd, "libstdc++", options, reference))
                            ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                 cppstd, "libc++", options, reference))
                        else:
                            ret.append(get_build("clang", arch, build_type_it, clang_version,
                                                 None, None, options, reference))
    return ret



import os
import textwrap

from conans import tools

from cpt.test.integration.base import BaseTest
from cpt.packager import ConanMultiPackager


class ProfileTest(BaseTest):

    def test_build_context_support(self):
        conanfile = textwrap.dedent("""
                from conans import ConanFile
                class Pkg(ConanFile):
                    pass
                """)
        self.save_conanfile(conanfile)
        build_profile = textwrap.dedent("""
                [settings]
                os=Linux
                arch=x86_64
                compiler=gcc
                compiler.version=8
                compiler.libcxx=libstdc++
                build_type=Release
                [options]
                [build_requires]
                [env]
                [conf]
                """)
        build_profile_path = os.path.join(self.tmp_folder, "build_profile")
        tools.save(build_profile_path, build_profile)

        host_profile = textwrap.dedent("""
                        [settings]
                        arch=armv7hf
                        build_type=Release
                        [options]
                        [build_requires]
                        [env]
                        [conf]
                        """)
        host_profile_path = os.path.join(self.tmp_folder, "host_profile")
        tools.save(host_profile_path, host_profile)
        self.packager = ConanMultiPackager(username="elcidcampeador",
                                           reference="tizona/1.0.40@elcidcampeador/testing", out=self.output.write)
        self.packager.add_common_builds()
        self.packager.run_builds(curpage=1, total_pages=1, base_profile_name=host_profile_path,
                                 base_profile_build_name=build_profile_path)
        self.assertIn("Using specified default base profile", self.output)
        self.assertIn("Using specified build profile", self.output)



import unittest
import os
import zipfile
import textwrap

from conans.client.tools import environment_append
from conans.model.manifest import FileTreeManifest
from conans.model.ref import PackageReference

from cpt.test.utils.tools import TestClient, TestServer
from cpt.test.unit.utils import MockCIManager

from cpt.test.test_client.tools import get_patched_multipackager


class UploadTest(unittest.TestCase):

    conanfile = """from conans import ConanFile
class Pkg(ConanFile):
    name = "lib"
    version = "1.0"
    options = {"shared": [True, False]}
    default_options = "shared=False"

    def build(self):
        self.output.warn("HALLO")
"""

    def setUp(self):
        for var in ["CONAN_USERNAME", "CONAN_CHANNEL", "CONAN_REFERENCE"]:
            try:
                del os.environ[var]
            except:
                pass
        self._ci_manager = MockCIManager()

    def test_dont_upload_non_built_packages(self):

        ts = TestServer(users={"user": "password"})
        tc = TestClient(servers={"default": ts}, users={"default": [("user", "password")]})
        tc.save({"conanfile.py": self.conanfile})
        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "user",
                                 "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user"}):
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True)
            mulitpackager.add({}, {"shared": True})
            mulitpackager.add({}, {"shared": False})
            mulitpackager.run()
            self.assertIn("Uploading package 1/2", tc.out)
            self.assertIn("Uploading package 2/2", tc.out)

            # With the same cache and server try to rebuild them with policy missing
            mulitpackager = get_patched_multipackager(tc, build_policy="missing",
                                                      exclude_vcvars_precommand=True)
            mulitpackager.add({}, {"shared": True})
            mulitpackager.add({}, {"shared": False})
            mulitpackager.run()
            self.assertIn("Skipping upload for 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9", tc.out)
            self.assertIn("Skipping upload for 2a623e3082a38f90cd2c3d12081161412de331b0", tc.out)
            self.assertNotIn("HALLO", tc.out)

            # Without any build policy they get built
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True)
            mulitpackager.add({}, {"shared": True})
            mulitpackager.add({}, {"shared": False})
            mulitpackager.run()
            self.assertNotIn("Skipping upload for 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9", tc.out)
            self.assertNotIn("Skipping upload for 2a623e3082a38f90cd2c3d12081161412de331b0", tc.out)
            self.assertIn("Uploading package 1/2", tc.out)
            self.assertIn("Uploading package 2/2", tc.out)
            self.assertIn("HALLO", tc.out)

    def test_upload_when_tag_is_false(self):
        ts = TestServer(users={"user": "password"})
        tc = TestClient(servers={"default": ts}, users={"default": [("user", "password")]})
        tc.save({"conanfile.py": self.conanfile})

        zip_path = os.path.join(tc.current_folder, 'config.zip')
        zipf = zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED)
        zipf.close()

        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "user",
                                 "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                 "CONAN_CONFIG_URL": zip_path, "CONAN_UPLOAD_ONLY_WHEN_TAG": "1",
                                 "TRAVIS": "1"}):

            mp = get_patched_multipackager(tc, exclude_vcvars_precommand=True)
            mp.add_common_builds(shared_option_name=False)
            mp.run()

            self.assertNotIn("Redefined channel by branch tag", tc.out)
            self.assertNotIn("Uploading packages for 'lib/1.0@user/stable'", tc.out)
            self.assertNotIn("Uploading package 1/1: 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9 to 'default'", tc.out)
            self.assertIn("Skipping upload, not tag branch", tc.out)

    def test_upload_when_tag_is_true(self):
        ts = TestServer(users={"user": "password"})
        tc = TestClient(servers={"default": ts}, users={"default": [("user", "password")]})
        tc.save({"conanfile.py": self.conanfile})

        zip_path = os.path.join(tc.current_folder, 'config.zip')
        zipf = zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED)
        zipf.close()

        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "user",
                                 "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                 "CONAN_CONFIG_URL": zip_path, "CONAN_UPLOAD_ONLY_WHEN_TAG": "1",
                                 "TRAVIS": "1", "TRAVIS_TAG": "0.1"}):

            mp = get_patched_multipackager(tc, exclude_vcvars_precommand=True)
            mp.add_common_builds(shared_option_name=False)
            mp.run()

            self.assertNotIn("Skipping upload, not tag branch", tc.out)
            self.assertIn("Redefined channel by branch tag", tc.out)
            self.assertIn("Uploading packages for 'lib/1.0@user/stable'", tc.out)
            self.assertIn("Uploading package 1/1: 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9 to 'default'", tc.out)

    def test_upload_only_recipe_env_var(self):
        ts = TestServer(users={"user": "password"})
        tc = TestClient(servers={"default": ts}, users={"default": [("user", "password")]})
        tc.save({"conanfile.py": self.conanfile})

        # Upload only the recipe
        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "user",
                                "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                "CONAN_UPLOAD_ONLY_RECIPE": "TRUE", "CONAN_CHANNEL": "mychannel"}):
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True,
                                                      ci_manager=self._ci_manager)
            mulitpackager.add({}, {"shared": True})
            mulitpackager.add({}, {"shared": False})
            mulitpackager.run()

            self.assertIn(" Uploading packages for 'lib/1.0@user/mychannel'", tc.out)
            self.assertIn("Uploading lib/1.0@user/mychannel to remote", tc.out)
            self.assertIn("Uploaded conan recipe 'lib/1.0@user/mychannel'", tc.out)
            self.assertNotIn("Uploading package 1/2", tc.out)
            self.assertNotIn("Uploading package 2/2", tc.out)

        # Re-use cache the upload the binary packages
        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "user",
                                "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                "CONAN_UPLOAD_ONLY_RECIPE": "FALSE", "CONAN_CHANNEL": "mychannel"}):
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True,
                                                      ci_manager=self._ci_manager)
            mulitpackager.add({}, {"shared": True})
            mulitpackager.add({}, {"shared": False})
            mulitpackager.run()

            self.assertIn(" Uploading packages for 'lib/1.0@user/mychannel'", tc.out)
            self.assertIn("Uploading lib/1.0@user/mychannel to remote", tc.out)
            self.assertNotIn("Recipe is up to date, upload skipped", tc.out)
            self.assertIn("Uploading package 1/2", tc.out)
            self.assertIn("Uploading package 2/2", tc.out)

    def test_upload_only_recipe_params(self):
        ts = TestServer(users={"user": "password"})
        tc = TestClient(servers={"default": ts}, users={"default": [("user", "password")]})
        tc.save({"conanfile.py": self.conanfile})

        # Upload only the recipe
        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "user",
                                "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                "CONAN_CHANNEL": "mychannel"}):
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True,
                                                      upload_only_recipe=True,
                                                      ci_manager=self._ci_manager)
            mulitpackager.add({}, {"shared": True})
            mulitpackager.add({}, {"shared": False})
            mulitpackager.run()

            self.assertIn(" Uploading packages for 'lib/1.0@user/mychannel'", tc.out)
            self.assertIn("Uploading lib/1.0@user/mychannel to remote", tc.out)
            self.assertIn("Uploaded conan recipe 'lib/1.0@user/mychannel'", tc.out)
            self.assertNotIn("Uploading package 1/2", tc.out)
            self.assertNotIn("Uploading package 2/2", tc.out)

        # Re-use cache the upload the binary packages
        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "user",
                                "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                "CONAN_CHANNEL": "mychannel"}):
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True,
                                                      upload_only_recipe=False,
                                                      ci_manager=self._ci_manager)
            mulitpackager.add({}, {"shared": True})
            mulitpackager.add({}, {"shared": False})
            mulitpackager.run()

            self.assertIn(" Uploading packages for 'lib/1.0@user/mychannel'", tc.out)
            self.assertIn("Uploading lib/1.0@user/mychannel to remote", tc.out)
            self.assertNotIn("Recipe is up to date, upload skipped", tc.out)
            self.assertIn("Uploading package 1/2", tc.out)
            self.assertIn("Uploading package 2/2", tc.out)

    def test_upload_package_revisions(self):
        ts = TestServer(users={"user": "password"})
        tc = TestClient(servers={"default": ts}, users={"default": [("user", "password")]})
        tc.save({"conanfile.py": self.conanfile})
        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "user",
                                 "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                 "CONAN_REVISIONS_ENABLED": "1"}):
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True,
                                                      ci_manager=self._ci_manager)
            mulitpackager.add({}, {"shared": True})
            mulitpackager.add({}, {"shared": False})
            mulitpackager.run()
            self.assertNotIn("Skipping upload for 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9", tc.out)
            self.assertNotIn("Skipping upload for 2a623e3082a38f90cd2c3d12081161412de331b0", tc.out)
            self.assertIn("Uploading package 1/2", tc.out)
            self.assertIn("Uploading package 2/2", tc.out)
            self.assertIn("HALLO", tc.out)

    def test_upload_partial_reference(self):
        ts = TestServer(users={"user": "password"},
                        write_permissions=[("lib/1.0", "user")])
        tc = TestClient(servers={"default": ts}, users={"default": [("user", "password")]})
        tc.save({"conanfile.py": self.conanfile})

        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "user",
                                 "CONAN_PASSWORD": "password", "CONAN_REFERENCE": "lib/1.0@",
                                 }):

            mp = get_patched_multipackager(tc, exclude_vcvars_precommand=True)
            mp.add_common_builds(shared_option_name=False)
            mp.run()

            self.assertIn("Uploading packages for 'lib/1.0@'", tc.out)
            self.assertIn("lib/1.0: WARN: HALLO", tc.out)

    def test_forced_upload(self):
        NO_SETTINGS_PACKAGE_ID = "5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9"
        conanfile = textwrap.dedent("""
                from conans import ConanFile
                class Pkg(ConanFile):

                    def configure(self):
                        self.output.warn("Mens sana in corpore sano")
        """)
        ts = TestServer(users={"foo": "password"})
        tc = TestClient(servers={"foo_server": ts}, users={"foo_server": [("foo", "password")]})

        tc.save({"conanfile.py": conanfile})
        tc.run("create . lib/1.0@foo/stable")
        tc.run("upload lib/1.0@foo/stable --all -r foo_server")
        pref = PackageReference.loads("lib/1.0@foo/stable#{}:{}".format(0, NO_SETTINGS_PACKAGE_ID))
        path = os.path.join(ts.server_store.package_revisions_root(pref), "0")
        manifest = FileTreeManifest.load(path)
        manifest.time += 1000
        manifest.save(path)

        tc.save({"conanfile.py": conanfile.replace("warn", "info")})
        tc.run("create . lib/1.0@foo/stable")

        # Force is True, package must be uploaded all times
        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "foo",
                                "CONAN_PASSWORD": "password", "CONAN_USERNAME": "foo",
                                "CONAN_UPLOAD_FORCE": "True"}):
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True)
            mulitpackager.add_common_builds(reference="lib/1.0@foo/stable",
                                            shared_option_name=False)
            mulitpackager.run()
            self.assertIn("Uploading packages for 'lib/1.0@foo/stable'", tc.out)
            self.assertNotIn("Recipe is up to date, upload skipped", tc.out)
            self.assertNotIn("Package is up to date, upload skipped", tc.out)

        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "foo",
                                "CONAN_PASSWORD": "password", "CONAN_USERNAME": "foo",
                                "CONAN_UPLOAD_FORCE": "FALSE"}):
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True,
                                                      upload_force=True)
            mulitpackager.add_common_builds(reference="lib/1.0@foo/stable",
                                            shared_option_name=False)
            mulitpackager.run()
            self.assertIn("Uploading packages for 'lib/1.0@foo/stable'", tc.out)
            self.assertNotIn("Package is up to date, upload skipped", tc.out)
            self.assertNotIn("Recipe is up to date, upload skipped", tc.out)

        # Force is False. Must not upload any package
        with environment_append({"CONAN_UPLOAD": ts.fake_url, "CONAN_LOGIN_USERNAME": "foo",
                                "CONAN_PASSWORD": "password", "CONAN_USERNAME": "foo",
                                "CONAN_UPLOAD_FORCE": "FALSE"}):
            mulitpackager = get_patched_multipackager(tc, exclude_vcvars_precommand=True,
                                                      upload_force=False)
            mulitpackager.add_common_builds(reference="lib/1.0@foo/stable",
                                            shared_option_name=False)
            mulitpackager.run()
            self.assertIn("Uploading packages for 'lib/1.0@foo/stable'", tc.out)
            self.assertIn("Recipe is up to date, upload skipped", tc.out)
            self.assertIn("Package is up to date, upload skipped", tc.out)


class UploadDependenciesTest(unittest.TestCase):

    conanfile_bar = """from conans import ConanFile
class Pkg(ConanFile):
    name = "bar"
    version = "0.1.0"

    def build(self):
        pass
    """

    conanfile_foo = """from conans import ConanFile
class Pkg(ConanFile):
    name = "foo"
    version = "1.0.0"
    options = {"shared": [True, False]}
    default_options = "shared=True"

    def build(self):
        pass
    """

    conanfile = """from conans import ConanFile
class Pkg(ConanFile):
    name = "foobar"
    version = "2.0"
    requires = "bar/0.1.0@foo/stable", "foo/1.0.0@bar/testing"

    def build(self):
        self.output.warn("BUILDING")
"""

    def setUp(self):
        self._ci_manager = MockCIManager()
        self._server = TestServer(users={"user": "password"},
                                  write_permissions=[("bar/0.1.0@foo/stable", "user"),
                                                     ("foo/1.0.0@bar/testing", "user")])
        self._client = TestClient(servers={"default": self._server},
                                 users={"default": [("user", "password")]})
        self._client.save({"conanfile_bar.py": self.conanfile_bar})
        self._client.run("export conanfile_bar.py foo/stable")
        self._client.save({"conanfile_foo.py": self.conanfile_foo})
        self._client.run("export conanfile_foo.py bar/testing")
        self._client.save({"conanfile.py": self.conanfile})

    def test_upload_all_dependencies(self):
        with environment_append({"CONAN_UPLOAD":  self._server.fake_url,
                                 "CONAN_LOGIN_USERNAME": "user",
                                 "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                 "CONAN_UPLOAD_DEPENDENCIES": "all"}):

            mulitpackager = get_patched_multipackager(self._client, username="user",
                                                      channel="testing",
                                                      build_policy="missing",
                                                      exclude_vcvars_precommand=True,
                                                      ci_manager=self._ci_manager)
            mulitpackager.add({}, {})
            mulitpackager.run()

            self.assertIn("Uploading packages for 'foobar/2.0@user/testing'", self._client.out)
            self.assertIn("Uploaded conan recipe 'foobar/2.0@user/testing'", self._client.out)
            self.assertIn("Uploading package 1/1: f88b82969cca9c4bf43f9effe1157e641f38f16d", self._client.out)

            self.assertIn("Uploading packages for 'bar/0.1.0@foo/stable'", self._client.out)
            self.assertIn("Uploaded conan recipe 'bar/0.1.0@foo/stable'", self._client.out)
            self.assertIn("Uploading package 1/1: 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9", self._client.out)

            self.assertIn("Uploading packages for 'foo/1.0.0@bar/testing'", self._client.out)
            self.assertIn("Uploaded conan recipe 'foo/1.0.0@bar/testing'", self._client.out)
            self.assertIn("Uploading package 1/1: 2a623e3082a38f90cd2c3d12081161412de331b0", self._client.out)

    def test_invalid_upload_dependencies(self):
        with environment_append({"CONAN_UPLOAD":  self._server.fake_url,
                                 "CONAN_LOGIN_USERNAME": "user",
                                 "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                 "CONAN_UPLOAD_DEPENDENCIES": "all,bar/0.1.0@foo/stable"}):
            with self.assertRaises(Exception) as context:
                get_patched_multipackager(self._client, exclude_vcvars_precommand=True)
            self.assertIn("Upload dependencies only accepts or 'all' or package references. Do not mix both!", str(context.exception))


    def test_upload_specific_dependencies(self):
        with environment_append({"CONAN_UPLOAD":  self._server.fake_url,
                                 "CONAN_LOGIN_USERNAME": "user",
                                 "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                 "CONAN_UPLOAD_DEPENDENCIES": "foo/1.0.0@bar/testing"}):

            mulitpackager = get_patched_multipackager(self._client, username="user",
                                                      channel="testing",
                                                      build_policy="missing",
                                                      exclude_vcvars_precommand=True,
                                                      ci_manager=self._ci_manager)
            mulitpackager.add({}, {})
            mulitpackager.run()

            self.assertIn("Uploading packages for 'foobar/2.0@user/testing'", self._client.out)
            self.assertIn("Uploaded conan recipe 'foobar/2.0@user/testing'", self._client.out)
            self.assertIn("Uploading package 1/1: f88b82969cca9c4bf43f9effe1157e641f38f16d", self._client.out)

            self.assertNotIn("Uploading packages for 'bar/0.1.0@foo/stable'", self._client.out)
            self.assertNotIn("Uploaded conan recipe 'bar/0.1.0@foo/stable'", self._client.out)
            self.assertNotIn("Uploading package 1/1: 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9", self._client.out)

            self.assertIn("Uploading packages for 'foo/1.0.0@bar/testing'", self._client.out)
            self.assertIn("Uploaded conan recipe 'foo/1.0.0@bar/testing'", self._client.out)
            self.assertIn("Uploading package 1/1: 2a623e3082a38f90cd2c3d12081161412de331b0", self._client.out)

    def test_upload_regex_dependencies(self):
        with environment_append({"CONAN_UPLOAD":  self._server.fake_url,
                                 "CONAN_LOGIN_USERNAME": "user",
                                 "CONAN_PASSWORD": "password", "CONAN_USERNAME": "user",
                                 "CONAN_UPLOAD_DEPENDENCIES": "foo/*"}):

            mulitpackager = get_patched_multipackager(self._client, username="user",
                                                      channel="testing",
                                                      build_policy="missing",
                                                      exclude_vcvars_precommand=True,
                                                      ci_manager=self._ci_manager)

            mulitpackager.add({}, {})
            mulitpackager.run()

            self.assertIn("Uploading packages for 'foobar/2.0@user/testing'", self._client.out)
            self.assertIn("Uploaded conan recipe 'foobar/2.0@user/testing'", self._client.out)
            self.assertIn("Uploading package 1/1: f88b82969cca9c4bf43f9effe1157e641f38f16d", self._client.out)

            self.assertNotIn("Uploading packages for 'bar/0.1.0@foo/stable'", self._client.out)
            self.assertNotIn("Uploaded conan recipe 'bar/0.1.0@foo/stable'", self._client.out)
            self.assertNotIn("Uploading package 1/1: 5ab84d6acfe1f23c4fae0ab88f26e3a396351ac9", self._client.out)

            self.assertNotIn("Uploading packages for 'foo/1.0.0@bar/testing'", self._client.out)
            self.assertNotIn("Uploaded conan recipe 'foo/1.0.0@bar/testing'", self._client.out)
            self.assertNotIn("Uploading package 1/1: 2a623e3082a38f90cd2c3d12081161412de331b0", self._client.out)



from conans.model.ref import ConanFileReference


class GenConanfile(object):
    """
    USAGE:

    x = GenConanfile().with_import("import os").\
        with_setting("os").\
        with_option("shared", [True, False]).\
        with_default_option("shared", True).\
        with_build_msg("holaaa").\
        with_build_msg("adiooos").\
        with_package_file("file.txt", "hola").\
        with_package_file("file2.txt", "hola")
    """

    def __init__(self, name=None, version=None):
        self._imports = ["from conans import ConanFile"]
        self._name = name
        self._version = version
        self._settings = []
        self._options = {}
        self._generators = []
        self._default_options = {}
        self._provides = []
        self._deprecated = None
        self._package_files = {}
        self._package_files_env = {}
        self._package_files_link = {}
        self._build_messages = []
        self._scm = {}
        self._requires = []
        self._requirements = []
        self._build_requires = []
        self._build_requirements = []
        self._revision_mode = None
        self._package_info = {}
        self._package_id_lines = []
        self._test_lines = []

    def with_name(self, name):
        self._name = name
        return self

    def with_version(self, version):
        self._version = version
        return self

    def with_provides(self, provides):
        self._provides.append(provides)
        return self

    def with_deprecated(self, deprecated):
        self._deprecated = deprecated
        return self

    def with_revision_mode(self, revision_mode):
        self._revision_mode = revision_mode
        return self

    def with_scm(self, scm):
        self._scm = scm
        return self

    def with_generator(self, generator):
        self._generators.append(generator)
        return self

    def with_require(self, ref, private=False, override=False):
        ref_str = ref.full_str() if isinstance(ref, ConanFileReference) else ref
        self._requires.append((ref_str, private, override))
        return self

    def with_requires(self, *refs):
        for ref in refs:
            self.with_require(ref)
        return self

    def with_requirement(self, ref, private=False, override=False):
        ref_str = ref.full_str() if isinstance(ref, ConanFileReference) else ref
        self._requirements.append((ref_str, private, override))
        return self

    def with_build_requires(self, *refs):
        for ref in refs:
            ref_str = ref.full_str() if isinstance(ref, ConanFileReference) else ref
            self._build_requires.append(ref_str)
        return self

    def with_build_requirement(self, ref, force_host_context=False):
        ref_str = ref.full_str() if isinstance(ref, ConanFileReference) else ref
        self._build_requirements.append((ref_str, force_host_context))
        return self

    def with_import(self, i):
        if i not in self._imports:
            self._imports.append(i)
        return self

    def with_setting(self, setting):
        self._settings.append(setting)
        return self

    def with_settings(self, *settings):
        self._settings.extend(settings)
        return self

    def with_option(self, option_name, values):
        self._options[option_name] = values
        return self

    def with_default_option(self, option_name, value):
        self._default_options[option_name] = value
        return self

    def with_package_file(self, file_name, contents=None, env_var=None, link=None):
        if not contents and not env_var:
            raise Exception("Specify contents or env_var")
        self.with_import("import os")
        self.with_import("from conans import tools")
        if contents:
            self._package_files[file_name] = contents
        if link:
            self._package_files_link[file_name] = link
        if env_var:
            self._package_files_env[file_name] = env_var
        return self

    def with_build_msg(self, msg):
        self._build_messages.append(msg)
        return self

    def with_package_info(self, cpp_info=None, env_info=None):
        assert isinstance(cpp_info, dict), "cpp_info ({}) expects dict".format(type(cpp_info))
        assert isinstance(env_info, dict), "env_info ({}) expects dict".format(type(env_info))
        if cpp_info:
            self._package_info["cpp_info"] = cpp_info
        if env_info:
            self._package_info["env_info"] = env_info
        return self

    def with_package_id(self, line):
        self._package_id_lines.append(line)
        return self

    def with_test(self, line):
        self._test_lines.append(line)
        return self

    @property
    def _name_line(self):
        if not self._name:
            return ""
        return "name = '{}'".format(self._name)

    @property
    def _version_line(self):
        if not self._version:
            return ""
        return "version = '{}'".format(self._version)

    @property
    def _provides_line(self):
        if not self._provides:
            return ""
        line = ", ".join('"{}"'.format(provide) for provide in self._provides)
        return "provides = {}".format(line)

    @property
    def _deprecated_line(self):
        if not self._deprecated:
            return ""
        return "deprecated = {}".format(self._deprecated)

    @property
    def _scm_line(self):
        if not self._scm:
            return ""
        line = ", ".join('"%s": "%s"' % (k, v) for k, v in self._scm.items())
        return "scm = {%s}" % line

    @property
    def _generators_line(self):
        if not self._generators:
            return ""
        line = ", ".join('"{}"'.format(generator) for generator in self._generators)
        return "generators = {}".format(line)

    @property
    def _revision_mode_line(self):
        if not self._revision_mode:
            return ""
        line = "revision_mode=\"{}\"".format(self._revision_mode)
        return line

    @property
    def _settings_line(self):
        if not self._settings:
            return ""
        line = ", ".join('"%s"' % s for s in self._settings)
        return "settings = {}".format(line)

    @property
    def _options_line(self):
        if not self._options:
            return ""
        line = ", ".join('"%s": %s' % (k, v) for k, v in self._options.items())
        tmp = "options = {%s}" % line
        return tmp

    @property
    def _default_options_line(self):
        if not self._default_options:
            return ""
        line = ", ".join('"%s": %s' % (k, v) for k, v in self._default_options.items())
        tmp = "default_options = {%s}" % line
        return tmp

    @property
    def _build_requirements_method(self):
        if not self._build_requirements:
            return ""

        lines = []
        for ref, force_host_context in self._build_requirements:
            force_host = ", force_host_context=True" if force_host_context else ""
            lines.append('        self.build_requires("{}"{})'.format(ref, force_host))
        return "def build_requirements(self):\n{}\n".format("\n".join(lines))

    @property
    def _build_requires_line(self):
        if not self._build_requires:
            return ""
        line = ", ".join(['"{}"'.format(r) for r in self._build_requires])
        tmp = "build_requires = %s" % line
        return tmp

    @property
    def _requires_line(self):
        if not self._requires:
            return ""
        items = []
        for ref, private, override in self._requires:
            if private or override:
                private_str = ", 'private'" if private else ""
                override_str = ", 'override'" if override else ""
                items.append('("{}"{}{})'.format(ref, private_str, override_str))
            else:
                items.append('"{}"'.format(ref))
        tmp = "requires = ({}, )".format(", ".join(items))
        return tmp

    @property
    def _requirements_method(self):
        if not self._requirements:
            return ""

        lines = []
        for ref, private, override in self._requirements:
            private_str = ", private=True" if private else ""
            override_str = ", override=True" if override else ""
            lines.append('        self.requires("{}"{}{})'.format(ref, private_str, override_str))

        return """
    def requirements(self):
{}
        """.format("\n".join(lines))

    @property
    def _package_method(self):
        lines = []
        if self._package_files:
            lines = ['        tools.save(os.path.join(self.package_folder, "{}"), "{}")'
                     ''.format(key, value)
                     for key, value in self._package_files.items()]

        if self._package_files_env:
            lines.extend(['        tools.save(os.path.join(self.package_folder, "{}"), '
                          'os.getenv("{}"))'.format(key, value)
                          for key, value in self._package_files_env.items()])
        if self._package_files_link:
            lines.extend(['        with tools.chdir(os.path.dirname('
                          'os.path.join(self.package_folder, "{}"))):\n'
                          '            os.symlink(os.path.basename("{}"), '
                          'os.path.join(self.package_folder, "{}"))'.format(key, key, value)
                          for key, value in self._package_files_link.items()])

        if not lines:
            return ""
        return """
    def package(self):
{}
    """.format("\n".join(lines))

    @property
    def _build_method(self):
        if not self._build_messages:
            return ""
        lines = ['        self.output.warn("{}")'.format(m) for m in self._build_messages]
        return """
    def build(self):
{}
    """.format("\n".join(lines))

    @property
    def _package_info_method(self):
        if not self._package_info:
            return ""
        lines = []
        if "cpp_info" in self._package_info:
            for k, v in self._package_info["cpp_info"].items():
                if k == "components":
                    for comp_name, comp in v.items():
                        for comp_attr_name, comp_attr_value in comp.items():
                            lines.append('        self.cpp_info.components["{}"].{} = {}'.format(
                                comp_name, comp_attr_name, str(comp_attr_value)))
                else:
                    lines.append('        self.cpp_info.{} = {}'.format(k, str(v)))
        if "env_info" in self._package_info:
            for k, v in self._package_info["env_info"].items():
                lines.append('        self.env_info.{} = {}'.format(k, str(v)))

        return """
    def package_info(self):
{}
        """.format("\n".join(lines))

    @property
    def _package_id_method(self):
        if not self._package_id_lines:
            return ""
        lines = ['        {}'.format(line) for line in self._package_id_lines]
        return """
    def package_id(self):
{}
        """.format("\n".join(lines))

    @property
    def _test_method(self):
        if not self._test_lines:
            return ""
        lines = ['', '    def test(self):'] + ['        %s' % m for m in self._test_lines]
        return "\n".join(lines)

    def __repr__(self):
        ret = []
        ret.extend(self._imports)
        ret.append("class HelloConan(ConanFile):")
        if self._name_line:
            ret.append("    {}".format(self._name_line))
        if self._version_line:
            ret.append("    {}".format(self._version_line))
        if self._provides_line:
            ret.append("    {}".format(self._provides_line))
        if self._deprecated_line:
            ret.append("    {}".format(self._deprecated_line))
        if self._generators_line:
            ret.append("    {}".format(self._generators_line))
        if self._requires_line:
            ret.append("    {}".format(self._requires_line))
        if self._build_requires_line:
            ret.append("    {}".format(self._build_requires_line))
        if self._requirements_method:
            ret.append("    {}".format(self._requirements_method))
        if self._build_requirements_method:
            ret.append("    {}".format(self._build_requirements_method))
        if self._scm:
            ret.append("    {}".format(self._scm_line))
        if self._revision_mode_line:
            ret.append("    {}".format(self._revision_mode_line))
        if self._settings_line:
            ret.append("    {}".format(self._settings_line))
        if self._options_line:
            ret.append("    {}".format(self._options_line))
        if self._default_options_line:
            ret.append("    {}".format(self._default_options_line))
        if self._build_method:
            ret.append("    {}".format(self._build_method))
        if self._package_method:
            ret.append("    {}".format(self._package_method))
        if self._package_info_method:
            ret.append("    {}".format(self._package_info_method))
        if self._package_id_lines:
            ret.append("    {}".format(self._package_id_method))
        if self._test_method:
            ret.append("    {}".format(self._test_method))
        if len(ret) == 2:
            ret.append("    pass")
        return "\n".join(ret)
